#include <string>
#include <assert.h>
#include <QtCore/QString>
#include <QtCore/QMutexLocker>
#include <QtCore/QThreadPool>
#include <QDialog>
#include <QtGui/QWindowStateChangeEvent>
#include <QMessageBox>

#include "TDMonitor.h"
#include "ui_VersionDialog.h"
#include "TDMonitor.moc"

#define LF "\n"

template<class T>
T minT(const T& a, const T& b) { return (a<b) ? a : b; }
template<class T>
T maxT(const T& a, const T& b) { return (a>b) ? a : b; }


// PP100API End processing
long endPP100Api()
{
	// Disconnecting from PC that TDBridge is running.
	long nRet = g_Api.PP100_DisConnectServer(g_Api.ulHandle);
	if(PP100API_SUCCESS != nRet)
	{
		fprintf(stderr, "Disconnecting from PC that TDBridge is running failed.\n");
		return nRet;
	}

	// Release PP100 Series API.
	nRet = g_Api.PP100_Destroy();
	if(PP100API_SUCCESS != nRet)
	{
		fprintf(stderr, "Destroying PP-100 Series API failed.\n");
	}
	return nRet;
}


// QThreadPool will supports function object in Qt 5.15
// To support older Qt, implement by myself.
class MyRunnable : public QRunnable
{
private:
	std::function<void ()> functionToRun;
public:
	MyRunnable(std::function<void ()> functionToRun)
	{
		this->functionToRun = functionToRun;
	}
	void run()
	{
		this->functionToRun();
	}
};


TDMonitorEx::TDMonitorEx(
	const PP100_ENUM_PUBLISHER_INFO& sPublisherInfo,
	QMainWindow *parent)
{
	m_sPublisherInfo = sPublisherInfo;

	setupUi(this);

	// Screen update cycle setting (normal time: 5000msec, window minimization: stop)
	m_timer.reset(new QTimer(this));
	m_timer->setInterval(5000);
	m_timer->setSingleShot(false);

	// Itegrate 2 actions into one by using QActionGroup
	m_exclusivActionShowJobs.reset(new QActionGroup(this));
	m_exclusivActionShowJobs->setExclusive(true);
	actionShowUnfinishedJobs->setActionGroup(m_exclusivActionShowJobs.get());
	actionShowFinishedJobs->setActionGroup(m_exclusivActionShowJobs.get());

	QTreeWidgetItem *header = treeJobList->headerItem();
	header->setFlags(header->flags() ^ Qt::ItemIsDragEnabled);
	treeJobList->clear();

	switch(m_sPublisherInfo.ulConnectType)
	{
	default:
	case PP100API_CONNECTTYPE_USB: // 0
		treeJobList->setColumnWidth(colJobID, 240);
		break;
	case PP100API_CONNECTTYPE_ETH: // 1
		treeJobList->setColumnWidth(colJobID, 240-80);
		header->setText(colJobPublisher=2, tr("Publisher", "person"));

		colUnfinishedCopies     = 3; // "Copies"
		colUnfinishedJobType    = 4; // "Type"
		colUnfinishedStackerIn  = 5; // "Input Stacker"
		colUnfinishedStackerOut = 6; // "Output Stacker"
		colUnfinishedRemaining  = 7; // "Time Needed"

		colFinishedComletionNumber    = 3; // "Successful"
		colFinishedErrorNumber        = 4; // "Failed"
		colFinishedPublicationNumber  = 5; // "Quantity"
		colFinishedJobType            = 6; // "Type"
		colFinishedStackerOut         = 7; // "Output Stacker"

		// to repaint column header
		showUnfinishedJobs();
	}

	// SIGNAL,SLOT settings
	QObject::connect(actionCancelJob, SIGNAL(triggered()), this, SLOT(cancelJob()));
	QObject::connect(actionSelectAllJobs, SIGNAL(triggered()), this, SLOT(selectAllJobs()));
	QObject::connect(actionShowUnfinishedJobs, SIGNAL(triggered()), this, SLOT(showUnfinishedJobs()));
	QObject::connect(actionShowFinishedJobs, SIGNAL(triggered()), this, SLOT(showFinishedJobs()));
	QObject::connect(actionShowAbout, SIGNAL(triggered()), this, SLOT(showAboutDialog()));
	QObject::connect(tabWidget, SIGNAL(currentChanged(int)), this, SLOT(changeTab(int)));
	QObject::connect(m_timer.get(), SIGNAL(timeout()), this, SLOT(updateStatusAsync()));
	QObject::connect(treeJobList, SIGNAL(clicked(QModelIndex)), this, SLOT(updateMenu()));

	// SIGNAL from fetchStatus()
	QObject::connect(this, SIGNAL(updateScreenSignal()), this, SLOT(updateScreen()));

	// Set context menu for TreeWidget
	treeJobList->addAction(actionCancelJob);
	treeJobList->addAction(actionSelectAllJobs);

	// Add Publisher name to window title
	QString title = windowTitle() + m_sPublisherInfo.pPublisherName;
	setWindowTitle(title);


	// calc minimum size of qlabel to hold content.
	int contentWidth = 80;
	contentWidth = maxT(contentWidth, labelStacker1Status->fontMetrics().boundingRect(tr("Input Stacker")).width());
	contentWidth = maxT(contentWidth, labelStacker1Status->fontMetrics().boundingRect(tr("Output Stacker")).width());
	contentWidth = maxT(contentWidth, labelStacker1Status->fontMetrics().boundingRect(tr("DVD(+R/-R) DL")).width());
	contentWidth = maxT(contentWidth, labelStacker1Status->fontMetrics().boundingRect(tr("BD-R XL 128")).width());
	labelStacker1Status->setMinimumSize(QSize(contentWidth + 10, 0));
	labelStacker2Status->setMinimumSize(QSize(contentWidth + 10, 0));
	labelStacker3Status->setMinimumSize(QSize(contentWidth + 10, 0));
	labelStacker4Status->setMinimumSize(QSize(contentWidth + 10, 0));

	contentWidth = 160;

	contentWidth = maxT(contentWidth, labelMode->fontMetrics().boundingRect(tr("External Output / Low-volume mode")).width());
	contentWidth = maxT(contentWidth, labelMode->fontMetrics().boundingRect(tr("Batch mode")).width());
	contentWidth = maxT(contentWidth, labelMode->fontMetrics().boundingRect(tr("Standard mode")).width());
	contentWidth = maxT(contentWidth, labelMode->fontMetrics().boundingRect(tr("Unknown")).width());
	labelMode->setMinimumSize(QSize(contentWidth + 10, 0));

	adjustSize();

	updateStatusAsync(); // -> updateScreenSignal() -> updateScreen()
	m_timer->start();
}


TDMonitorEx::~TDMonitorEx()
{
	if( m_timer )
	{
		m_timer->stop();
	}

	if ( m_mtxFetchStatus.tryLock(60*1000) )
	{
		m_mtxFetchStatus.unlock();
	}
	if ( m_mtxUpdateScreen.tryLock(60*1000) )
	{
		m_mtxUpdateScreen.unlock();
	}

	endPP100Api();
}


// Screen information update
void TDMonitorEx::fetchStatus(void)
{
	QMutexLocker locker(&m_mtxFetchStatus);

	std::deque<PP100_JOB_STATUS_EX> jobStatusPublishing;
	std::deque<PP100_JOB_STATUS_EX> jobStatusFinished;
	auto classify = [&](PP100_JOB_STATUS_EX& jobStatus)
	{
		if (jobStatus.pJobID[0] == '\0')
		{
			return;
		}
		if (
			jobStatus.ulJobStatus != PP100API_JOB_PUBLISHED &&
			jobStatus.ulJobStatus != PP100API_JOB_USER_CANCELLED &&
			jobStatus.ulJobStatus != PP100API_JOB_ERROR_CANCELLED &&
			jobStatus.ulJobStatus != PP100API_JOB_PUBLISHED_WARNING_NO_ERROR_DISC &&
			jobStatus.ulJobStatus != PP100API_JOB_PUBLISHED_WARNING_ERROR_DISC )
		{
			jobStatusPublishing.push_back(jobStatus);
		}
		else
		{
			jobStatusFinished.push_front(jobStatus);
		}
	};

	if ( g_Api.PP100_GetJobStatusEx != nullptr )
	{
		// ver 2.00
		unsigned long ulJobStatusNum = 0;
		long nRet = g_Api.PP100_GetJobStatusEx(g_Api.ulHandle, NULL, m_sPublisherInfo.pPublisherName, NULL, &ulJobStatusNum);
		// Job list creation
		if(PP100API_MORE_ITEMS == nRet)
		{
			while(PP100API_MORE_ITEMS == nRet)
			{
				std::vector<PP100_JOB_STATUS_EX> jobStatusWhole(ulJobStatusNum);
				nRet = g_Api.PP100_GetJobStatusEx(g_Api.ulHandle, NULL, m_sPublisherInfo.pPublisherName, &jobStatusWhole[0], &ulJobStatusNum);
				if(PP100API_SUCCESS == nRet)
				{
					for(auto& jobStatus : jobStatusWhole)
					{
						classify(jobStatus);
					}
				}
			}
		}
		else if (PP100API_SUCCESS != nRet)
		{
			fprintf(stderr, "Getting PP-100 Series JOB Status failed.\n");
		}
	}
	else
	{
		// ver 1.0x
		unsigned long ulJobStatusNum = 0;
		long nRet = g_Api.PP100_GetJobStatus(g_Api.ulHandle, NULL, NULL, &ulJobStatusNum);
		// Job list creation
		if(PP100API_MORE_ITEMS == nRet)
		{
			while(PP100API_MORE_ITEMS == nRet)
			{
				std::vector<PP100_JOB_STATUS> jobStatusWhole(ulJobStatusNum);
				nRet = g_Api.PP100_GetJobStatus(g_Api.ulHandle, NULL, &jobStatusWhole[0], &ulJobStatusNum);
				if(PP100API_SUCCESS == nRet)
				{
					for(auto& jobStatusV1 : jobStatusWhole)
					{
						if(strcmp(jobStatusV1.pPublisherName, m_sPublisherInfo.pPublisherName) == 0)
						{
							PP100_JOB_STATUS_EX jobStatus = {};

							strcpy(jobStatus.pJobID,         jobStatusV1.pJobID);
							strcpy(jobStatus.pPublisherName, jobStatusV1.pPublisherName);
							jobStatus.ulJobStatus                   = jobStatusV1.ulJobStatus;
							jobStatus.ulErrorCode                   = jobStatusV1.ulErrorCode;
							jobStatus.ulPublicationNumber           = jobStatusV1.ulPublicationNumber;
							jobStatus.ulCompletionNumber            = jobStatusV1.ulCompletionNumber;
							jobStatus.ulJobType                     = jobStatusV1.ulJobType;
							jobStatus.ulSource                      = jobStatusV1.ulSource;
							jobStatus.ulDestination                 = jobStatusV1.ulDestination;
							jobStatus.ulJobIndex                    = jobStatusV1.ulJobIndex;
							jobStatus.ulEstimateTime                = jobStatusV1.ulEstimateTime;
							jobStatus.ulErrorNumber                 = jobStatusV1.ulErrorNumber;
							jobStatus.ulMode                        = jobStatusV1.ulMode;
							jobStatus.ucIsSubmittedByTotalDiscMaker = jobStatusV1.ucIsSubmittedByTotalDiscMaker;
							jobStatus.ulRemainingEstimateTime       = jobStatusV1.ulRemainingEstimateTime;

							classify(jobStatus);
						}
					}
				}
			}
		}
		else if (PP100API_SUCCESS != nRet)
		{
			fprintf(stderr, "Getting PP-100 Series JOB Status failed.\n");
		}
	}

	// Device Status
	PP100_PUBLISHER_STATUS publisherStatus = {};
	long nRetGetPublisherStatus = g_Api.PP100_GetPublisherStatus(g_Api.ulHandle, m_sPublisherInfo.pPublisherName, &publisherStatus);

	{
		QMutexLocker locker(&m_mtxUpdateScreen);
		m_jobStatusUnfinished = jobStatusPublishing;
		m_jobStatusFinished = jobStatusFinished;

		m_nRetGetPublisherStatus = nRetGetPublisherStatus;
		m_publisherStatus = publisherStatus;
		m_bFirstGetPublisherStatusDone = true;
	}

	emit updateScreenSignal(); // -> updateScreen()
}


void TDMonitorEx::updateScreen_JobList(void)
{
	QMutexLocker locker(&m_mtxUpdateScreen);

	// Remember selected items.
	std::vector<std::string> jobIdList;
	for(auto item : treeJobList->selectedItems())
	{
		jobIdList.push_back(item->text(0).toStdString());
	}

	// Refresh list.
	treeJobList->clear();
	if(tabWidget->currentIndex() == 0)
	{
		for(auto & jobStatus : m_jobStatusUnfinished)
		{
			appendTreeItem(jobStatus, JobGroup::unfinished);
		}
	}
	else
	{
		for(auto & jobStatus : m_jobStatusFinished)
		{
			appendTreeItem(jobStatus, JobGroup::finished);
		}
	}

	// Restore selected items.
	if (!jobIdList.empty())
	{
		for (int i=0; i < treeJobList->topLevelItemCount(); i++)
		{
			QTreeWidgetItem * item = treeJobList->topLevelItem(i);
			for(auto jobId : jobIdList)
			{
				if ( item->text(0).toStdString() == jobId )
				{
					item->setSelected(true);
					break;
				}
			}
		}
	}
}


void TDMonitorEx::updateScreen_DeviceStatus(void)
{
	QMutexLocker locker(&m_mtxUpdateScreen);

	// Device Status
	if ( m_bFirstGetPublisherStatusDone == false )
	{
		return;
	}
	else if(PP100API_SUCCESS == m_nRetGetPublisherStatus)
	{
		// set text on textEdit_DeviceStatus based on the status
		setDeviceInfo(m_publisherStatus, m_jobStatusUnfinished);
	}
	else
	{
		textDeviceStatus->setText(
			tr("L7007:\n"
				"Communication with the specified output device has been interrupted.\n"
				"The communication settings may have been changed.\n"
				"Restart Total Disc Monitor."));
		return;
	}

	// Device Status
	// Ink Status
	auto setInkStatus = [&] (QProgressBar* barInk, unsigned long ulInk) -> void
	{
		if(
			m_publisherStatus.ulINFORMATIONCode[0] != PP100API_HNW000 &&
			ulInk >= 0 && ulInk <= 100)
		{
			barInk->setEnabled(true);
			if(ulInk != 0)
			{
				barInk->setValue(ulInk);
			}
			else
			{
				barInk->reset();
			}
		}
		else
		{
			barInk->setEnabled(false);
			barInk->reset();
		}
	};

	setInkStatus(barInkC,  m_publisherStatus.stInkStatus.ulCyan);
	setInkStatus(barInkLC, m_publisherStatus.stInkStatus.ulLightCyan);
	setInkStatus(barInkLM, m_publisherStatus.stInkStatus.ulLightMagenta);
	setInkStatus(barInkM,  m_publisherStatus.stInkStatus.ulMagenta);
	setInkStatus(barInkY,  m_publisherStatus.stInkStatus.ulYellow);
	setInkStatus(barInkK,  m_publisherStatus.stInkStatus.ulBlack);

	// Drive Status
	setDriveInfo(m_publisherStatus, 0);
	setDriveInfo(m_publisherStatus, 1);

	// Printer Status
	switch(m_publisherStatus.ulPrinterStatus)
	{
	case PP100API_TRANSFERRING:
		label_PrinterStatus->setText(tr("Transferring"));
		break;
	case PP100API_PRINTING:
		label_PrinterStatus->setText(tr("Printing"));
		break;
	case PP100API_PAUSED:
		label_PrinterStatus->setText(tr("Stopping"));
		break;
	case PP100API_CLEANING:
		label_PrinterStatus->setText(tr("Cleaning", "status"));
		break;
	case PP100API_DRYING:
		label_PrinterStatus->setText(tr("Drying Ink"));
		break;
	case PP100API_ERROR:
		label_PrinterStatus->setText(tr("Error"));
		break;
	case PP100API_WAITING:
		label_PrinterStatus->setText(tr("Standby"));
		break;
	default:
		fprintf(stderr, "unknown printer status: %lu\n", m_publisherStatus.ulPrinterStatus);
		label_PrinterStatus->setText(tr("Unknown"));
		break;
	}

	// Mode
	switch(m_publisherStatus.ulMode)
	{
	case PP100API_OUTPUT_MODE:
		labelMode->setText(tr("External Output / Low-volume mode"));
		break;
	case PP100API_BATCH_MODE:
		labelMode->setText(tr("Batch mode"));
		break;
	case PP100API_NORMAL_MODE:
		labelMode->setText(tr("Standard mode"));
		break;
	default:
		fprintf(stderr, "unknown publish mode: %lu\n", m_publisherStatus.ulMode);
		labelMode->setText(tr("Unknown"));
		break;
	}

	// Stacker status setting
	for(int nCnt = 0; nCnt < 4; ++nCnt)
	{
		setStackerSetting(m_publisherStatus, nCnt);
	}

	// Check if you are connected with USB3.0 and set the icon
	if ( m_sPublisherInfo.ulConnectType == PP100API_CONNECTTYPE_USB )
	{
		if(m_nRetGetPublisherStatus == PP100API_SUCCESS)
		{
			if(PP100API_USB_30 == m_publisherStatus.ulUsbConnectionMode)
			{
				iconUSB->setPixmap(QPixmap(QString::fromUtf8(":/res/E27-1.png")));
			}
			else
			{
				iconUSB->setPixmap(QPixmap(QString::fromUtf8(":/res/E27-2.png")));
			}
		}
		else
		{
			iconUSB->setPixmap(QPixmap(QString::fromUtf8(":/res/E27-2.png")));
		}
	}
	else
	{
		iconUSB->hide();
	}
}


// Add TreeWidgetItem with Job information
void TDMonitorEx::appendTreeItem(const PP100_JOB_STATUS_EX& jobStatus, JobGroup jobGroup)
{
	if (jobStatus.pJobID[0] == '\0')
	{
		return;
	}

	QTreeWidgetItem* item = new QTreeWidgetItem(treeJobList);

	// Job ID
	if (jobStatus.pJobName != NULL && jobStatus.pJobName[0] != '\0')
	{
		item->setText(colJobID, jobStatus.pJobName);
	}
	else
	{
		item->setText(colJobID, jobStatus.pJobID);
	}
	item->setData(colJobID, Qt::UserRole, QString(jobStatus.pJobID));

	// Job Publication Person
	if ( colJobPublisher >= 0 )
	{
		item->setText(colJobPublisher, jobStatus.pPublicationPerson);
	}

	// Status
	item->setText(colJobStatus, convJobStatus(jobStatus));

	if(jobGroup == JobGroup::unfinished)
	{
		// Publishing Job
		item->setTextAlignment(colUnfinishedCopies, Qt::AlignRight|Qt::AlignVCenter);
		item->setTextAlignment(colUnfinishedJobType, Qt::AlignLeft|Qt::AlignVCenter);
		item->setTextAlignment(colUnfinishedStackerIn, Qt::AlignLeft|Qt::AlignVCenter);

		// Copies
		if ( jobStatus.ulPublicationNumber == 0 )
		{
			item->setText(colUnfinishedCopies, "-");
		}
		else
		{
			item->setText(colUnfinishedCopies, QString::asprintf("%lu / %lu", jobStatus.ulCompletionNumber, jobStatus.ulPublicationNumber));
		}
		// Type
		item->setText(colUnfinishedJobType, convJobType(jobStatus));
		item->setData(colUnfinishedJobType, Qt::UserRole, QVariant( (qulonglong)jobStatus.ulJobType ) );
		// Input Stacker
		item->setText(colUnfinishedStackerIn, convStackerIn(jobStatus));
		// Output Stacker
		item->setText(colUnfinishedStackerOut, convStackerOut(jobStatus));
		// Time Needed
		item->setText(colUnfinishedRemaining, convRemainingEstimateTime(jobStatus));
	}
	else
	{
		// Finished Job
		item->setTextAlignment(colFinishedComletionNumber, Qt::AlignRight|Qt::AlignVCenter);
		item->setTextAlignment(colFinishedErrorNumber, Qt::AlignRight|Qt::AlignVCenter);
		item->setTextAlignment(colFinishedPublicationNumber, Qt::AlignRight|Qt::AlignVCenter);

		// Successful
		item->setText(colFinishedComletionNumber, QString::asprintf("%lu", jobStatus.ulCompletionNumber));

		// Failed
		item->setText(colFinishedErrorNumber, QString::asprintf("%lu", jobStatus.ulErrorNumber));
		// Quantity
		item->setText(colFinishedPublicationNumber, QString::asprintf("%lu", jobStatus.ulPublicationNumber));
		// Type
		item->setText(colFinishedJobType, convJobType(jobStatus));
		// Output Stacker
		item->setText(colFinishedStackerOut, convStackerOut(jobStatus));
	}
}


// TreeWidget type setting process
QString TDMonitorEx::convJobType(const PP100_JOB_STATUS_EX& jobStatus)
{
	switch(jobStatus.ulJobType)
	{
	case PP100API_JOB_TYPE_PUBLISH: return tr("Publish");
	case PP100API_JOB_TYPE_WRITE:   return tr("Write");
	case PP100API_JOB_TYPE_PRINT:   return tr("Print");
	case PP100API_JOB_TYPE_MEASURE: return tr("Measure");
	case PP100API_JOB_TYPE_CLEANING: return tr("Cleaning", "job type");
	case PP100API_JOB_TYPE_NOZZLECHECK: return tr("Nozzle Check");
	case PP100API_JOB_TYPE_GAPADJUSTMENT: return tr("Print Head Alignment");
	case PP100API_JOB_TYPE_POSITIONOFFSET: return tr("Printing Position Correction");
	default:
		fprintf(stderr, "unknown job type: %lu\n", jobStatus.ulJobType);
	case PP100API_ERROR: return tr("Unknown");
		break;
	}
}


// TreeWidget type setting process
QString TDMonitorEx::convJobStatus(const PP100_JOB_STATUS_EX& jobStatus)
{
	switch(jobStatus.ulJobStatus)
	{
	case PP100API_JOB_RUNNING:        return tr("Publishing");
	case PP100API_JOB_CANCELING:      return tr("Canceling");
	case PP100API_JOB_PAUSING:        return tr("Pausing");
	case PP100API_JOB_PAUSED:         return tr("Paused");
	case PP100API_JOB_USER_CANCELLED: return tr("Cancel");
	case PP100API_JOB_PUBLISHED:      return tr("Published");
	case PP100API_JOB_WAITING:        return tr("Waiting");
	case PP100API_JOB_RECOVERING:     return tr("Recovering");
	case PP100API_JOB_STANDBY:        return tr("Standby_job");
	case PP100API_JOB_PUBLISHED_WARNING_NO_ERROR_DISC:
	case PP100API_JOB_PUBLISHED_WARNING_ERROR_DISC:
		return tr("Published (Warning)");
		break;
	case PP100API_JOB_ERROR_CANCELLED:
		return tr("Error");
		break;
	default:
		fprintf(stderr, "unknown job status: %lu\n", jobStatus.ulJobStatus);
		return tr("Unknown");
		break;
	}
}


QString TDMonitorEx::convStackerIn(const PP100_JOB_STATUS_EX& jobStatus)
{
	switch(jobStatus.ulSource)
	{
	case PP100API_STACKER1: return "1";
	case PP100API_STACKER2: return "2";
	case PP100API_STACKER3: return "3";
	case PP100API_STACKER4: return "4";
	case PP100API_STACKER1_OR_STACKER2:
		if(PP100API_BATCH_MODE==jobStatus.ulMode)
		{
			return "2, 1";
		}
		else
		{
			return "1, 2";
		}
		break;

	case PP100API_STACKER2_OR_STACKER3:
		if(PP100API_BATCH_MODE==jobStatus.ulMode)
		{
			return "3, 2";
		}
		else
		{
			return "2, 3";
		}
		break;

	case PP100API_NOSETTING:
		return "-";

	default:
		fprintf(stderr, "unknown stacker src: %lu\n", jobStatus.ulSource);
		return tr("Unknown");
		break;
	}
}


QString TDMonitorEx::convStackerOut(const PP100_JOB_STATUS_EX& jobStatus)
{
	switch(jobStatus.ulDestination)
	{
	case PP100API_STACKER1: return "1";
	case PP100API_STACKER2: return "2";
	case PP100API_STACKER3: return "3";
	case PP100API_STACKER4: return "4";
	case PP100API_STACKER2_OR_STACKER3:
		if(PP100API_BATCH_MODE==jobStatus.ulMode)
		{
			return "3, 2";
		}
		else
		{
			return "2, 3";
		}
		break;

	case PP100API_STACKER2_OR_STACKER4:
		return "2, 4";
		break;

	case PP100API_NOSETTING:
		return "-";

	default:
		fprintf(stderr, "unknown stacker dst: %lu\n", jobStatus.ulDestination);
		return tr("Unknown");
		break;
	}
}


QString TDMonitorEx::convRemainingEstimateTime(const PP100_JOB_STATUS_EX& jobStatus)
{
	if ( jobStatus.ulRemainingEstimateTime == 0 )
	{
		return tr("-");
	}
	else
	{
		int minutes = (jobStatus.ulRemainingEstimateTime + 59) / 60;

		if(minutes > 60)
		{
			return tr("%1 hr %2 min").arg(minutes / 60).arg(minutes % 60);
		}
		else
		{
			return tr("%1 min").arg(minutes);
		}
	}
}


// Device state settings
void TDMonitorEx::setDeviceInfo(const PP100_PUBLISHER_STATUS& publisherStatus, const std::deque<PP100_JOB_STATUS_EX>& jobStatusPublishing)
{
	QString msg;

	textDeviceStatus->clear();
	for(unsigned long ul = 0; ul < _countof(publisherStatus.ulINFORMATIONCode) && publisherStatus.ulINFORMATIONCode[ul] != 0; ul++)
	{
		if(ul)
		{
			msg += "\n";
		}
		switch(publisherStatus.ulINFORMATIONCode[ul])
		{
		case PP100API_HPR000:
			// A printer fatal error has occurred.
			msg += tr("L7011:\n"
				"An internal printer error occurred.\n"
				"Turn off the output device, check inside the printer, and then turn on the output device again.\n"
				"If the same error occurs frequently, contact Customer Support.");
			break;

		case PP100API_HPR001:
		case PP100API_HPR006:
			// A printer maintenance error has occurred.
			msg += tr("L7010:\n"
				"A printer maintenance error occurred.\n"
				"For details, please contact support.");
			break;

		case PP100API_HPR002:
			// The waste ink pads need to be replaced.
			msg += tr("L7009:\n"
				"Ink Pad Full\n"
				"\n"
				"The ink pad for this device is full and the device can no longer function.\n"
				"Please consult your manual for the recommended action.");
			break;

		case PP100API_HPR003:
			// A communication with the printer is failed.
			msg += tr("L7016:\n"
				"Communication with the printer in the output device has been interrupted.\n"
				"The port settings for the printer driver may have been changed.\n"
				"Correct the port settings.");
			break;

		case PP100API_HPR005:
			// The ink cartridge cover is open.
			msg += tr("L7113:\n"
				"The ink cartridge cover is open.\n"
				"Close it.");
			break;

		case PP100API_HPR008:
			// Time to replace the maintenance box.
			msg += tr("L7601:\n"
				"It is time to replace the Maintenance Box. No further printing is possible.\n"
				"Please replace the Maintenance Box with a new one.\n"
				"\n"
				"Epson brand products are recommended.");
			break;

		case PP100API_HPR007:	// The maintenance box is not installed.
		case PP100API_HPR009:	// Cannot recognize the maintenance box.
			msg += tr("L7600:\n"
				"The Maintenance Box cannot be recognized.\n"
				"Set the Maintenance Box properly.\n"
				"\n"
				"Epson brand products are recommended.");
			break;

		case PP100API_HPR010:
			// The maintenance box cover is not placed correctly.
			msg += tr("L7602:\n"
				"The Maintenance Box cover is not set properly.\n"
				"Please check the cover.");
			break;

		case PP100API_HDR000:
			// Controlling the drive tray is failed.
			msg += tr("L7003:\n"
				"Opening and closing the drive tray failed.\n"
				"There may be a disc left inside the drive.\n"
				"Please turn off the output device and check inside the drive.\n"
				"If the same error occurs frequently, please contact support.");
			break;

		case PP100API_HDR001:
			// Processing the job is impossible because of a fatal error of the drive.
			msg += tr("L7013:\n"
				"An internal drive error occurred.\n"
				"Turn off the output device, unplug the power cable, and check inside the drive.\n"
				"If the same error occurs frequently, contact Customer Support.");
			break;

		case PP100API_HDR002:
			// The drive is not connected correctly.
			msg += tr("L7020:\n"
				"An error occured when communicating with the drive.\n"
				"Turn off the output device, unplug the power cable, and check the drive type and connection.\n"
				"If the same error occurs frequently, contact Customer Support.");
			break;

		case PP100API_HAR000:
			// Picking up a disc from the drive tray or the printer tray is failed.
			msg += tr("L7005:\n"
				"Disc pickup failed.\n"
				"There is a disc left in the drive or printer tray.\n"
				"Please turn off the output device and remove the disc from the tray.\n"
				"If the same error occurs frequently, please contact support.");
			break;

		case PP100API_HAR001:
			// A disc is dropped while being picked up.
			msg += tr("L7001:\n"
				"The arm dropped the disc while transferring it.\n"
				"Please turn off the output device and remove the disc from the equipment.\n"
				"If the same error occurs frequently,\n"
				"please contact support.");
			break;

		case PP100API_HAR002:
			// Releasing a disc is failed.
			msg += tr("L7006:\n"
				"Disc release failed.\n"
				"Please turn the output device off and then on again.\n"
				"If the same error occurs frequently, please contact support.");
			break;

		case PP100API_HAR003:
			// Setting a disc in the drive tray or the printer tray must have been made, however, no disc is detected in the tray.
			msg += tr("L7017:\n"
				"Detecting the transferred disc failed.\n"
				"Turn off the output device and remove the disc from the device.\n"
				"If the same error occurs frequently, contact Customer Support.");
			break;

		case PP100API_HAR004:
			// An error occurs while the robotic arm is moving.
			msg += tr("L7002:\n"
				"An error occurred while the arm was moving.\n"
				"There may be an obstruction in the path of the arm.\n"
				"Please turn off the output device and check inside the device.\n"
				"If the same error occurs frequently, please contact support.");
			break;

		case PP100API_HAR005:
			// An auto loader fatal error has occurred. The robotic arm can not be moved due to overheating of the motor.
			msg += tr("L7015:\n"
				"An internal auto loader error occurred.\n"
				"Turn off the output device, and check inside the device.\n"
				"If the same error occurs frequently, contact Customer Support.");
			break;

		case PP100API_HAR007:
			// Picking up a disc from the stacker is failed.
			msg += tr("L7114:\n"
				"Disc pickup failed.\n"
				"To ensure that discs are transferred properly, be sure to break the static attraction between discs in the stacker.\n"
				"If pickup fails after separating the discs, contact Customer Support.");
			break;

		case PP100API_HAR008:
			// Multiple discs have been transferred to the drive.
			msg += tr("L7104:\n"
				"Detected multiple discs transferred to the drive.\n"
				"Remove all discs from the drive tray to avoid causing any damage to the output device.\n"
				"Also, to ensure that discs are transferred properly, check that the discs are loosely arranged in the stacker.");
			break;

		case PP100API_HAR009:
			// Multiple discs have been transferred to the printer.
			msg += tr("L7105:\n"
				"Detected multiple discs transferred to the printer.\n"
				"Remove all discs from the printer tray to avoid causing any damage to the output device.\n"
				"Also, to ensure that discs are transferred properly, check that the discs are loosely arranged in the stacker.");
			break;

		case PP100API_HIKC000:	// The cyan ink cartridge has been removed.
		case PP100API_HIKC001:	// The model number of the cyan ink cartridge is different or the model number cannot be recognized.
			msg += tr("L7110:\n"
				"The following ink cartridge cannot be recognized.\n"
				"Set the ink cartridge correctly.\n"
				"\n"
				"Cyan (EPSON brand part number:PJIC1(C))\n"
				"Epson brand products are recommended.");
			break;

		case PP100API_HIKM000:	// The magenta ink cartridge has been removed.
		case PP100API_HIKM001:	// The model number of the magenta ink cartridge is different or the model number cannot be recognized.
			msg += tr("L7110:\n"
				"The following ink cartridge cannot be recognized.\n"
				"Set the ink cartridge correctly.\n"
				"\n"
				"Magenta (EPSON brand part number:PJIC4(M))\n"
				"Epson brand products are recommended.");
			break;

		case PP100API_HIKY000:	// The yellow ink cartridge has been removed.
		case PP100API_HIKY001:	// The model number of the yellow ink cartridge is different or the model number cannot be recognized.
			msg += tr("L7110:\n"
				"The following ink cartridge cannot be recognized.\n"
				"Set the ink cartridge correctly.\n"
				"\n"
				"Yellow (EPSON brand part number:PJIC5(Y))\n"
				"Epson brand products are recommended.");
			break;

		case PP100API_HIKLC000:	// The light cyan ink cartridge has been removed.
		case PP100API_HIKLC001:	// The model number of the light cyan ink cartridge is different or the model number cannot be recognized.
			msg += tr("L7110:\n"
				"The following ink cartridge cannot be recognized.\n"
				"Set the ink cartridge correctly.\n"
				"\n"
				"Light cyan (EPSON brand part number:PJIC2(LC))\n"
				"Epson brand products are recommended.");
			break;

		case PP100API_HIKLM000:	// The light magenta ink cartridge has been removed.
		case PP100API_HIKLM001:	// The model number of the light magenta ink cartridge is different or the model number cannot be recognized.
			msg += tr("L7110:\n"
				"The following ink cartridge cannot be recognized.\n"
				"Set the ink cartridge correctly.\n"
				"\n"
				"Light magenta (EPSON brand part number:PJIC3(LM))\n"
				"Epson brand products are recommended.");
			break;

		case PP100API_HIKBK000:	// The black ink cartridge has been removed.
		case PP100API_HIKBK001:	// The model number of the black ink cartridge is different or the model number cannot be recognized.
			msg += tr("L7110:\n"
				"The following ink cartridge cannot be recognized.\n"
				"Set the ink cartridge correctly.\n"
				"\n"
				"Black (EPSON brand part number:PJIC6(K))\n"
				"Epson brand products are recommended.");
			break;

		case PP100API_HST1000:
			// The number of discs in the stacker 1 has exceeded the upper limit.
			msg += tr("L7108：\n"
				"There are too many discs in Stacker 1.\n"
				"Please remove some discs.");
			break;

		case PP100API_HST1001:
			// The stacker 1 has not been installed.
			msg += tr("L7103:\n"
				"Stacker 1 is not set properly.\n"
				"Please check the stacker.");
			break;

		case PP100API_HST2000:
			// The number of discs in the stacker 2 has exceeded the upper limit.
			msg += tr("L7108：\n"
				"There are too many discs in Stacker 2.\n"
				"Please remove some discs.");
			break;

		case PP100API_HST2001:
			// The stacker 2 has not been installed.
			msg += tr("L7103:\n"
				"Stacker 2 is not set properly.\n"
				"Please check the stacker.");
			break;

		case PP100API_HST3000:
			// The number of discs in the stacker 3 has exceeded the upper limit.
			msg += tr("L7108：\n"
				"There are too many discs in Stacker 3.\n"
				"Please remove some discs.");
			break;

		case PP100API_HST3001:
			// The output stacker in Standard mode is set to stacker 3, but stacker 3 is not installed,
			// or the output stacker in External Output mode is set to stacker 3, but stacker 3 is not installed,
			// or stacker 3 is not installed in Batch mode.
			msg += tr("L7103:\n"
				"Stacker 3 is not set properly.\n"
				"Please check the stacker.");
			break;

		case PP100API_HST3002:
			// The output stacker in Standard mode is set to stacker 4, or the output stacker in External Output mode is set to stacker 4, but stacker 3 is installed.
			msg += tr("L7106:\n"
				"Stacker 3 is not necessary.\n"
				"Please remove Stacker 3.");
			break;

		case PP100API_HDC000:
			// The disc cover is open.
			msg += tr("L7102:\n"
				"The disc cover is open.\n"
				"Please close it.");
			break;

		case PP100API_HNW000:
			// Either one of the following communication errors occur.
			// (1) Communication with Disc producer is disabled.
			// (2) The filter driver has not been attached.
			msg += tr("L7000:\n"
				"An error occurred in communication with the output device.\n"
				"Please check the power supply of the output device and the connection.");
			break;

		case PP100API_HNW001:
			// A non-supported version error has occurred, and any of the following has been detected.
			// (1) Firmware version of the auto loader is not supported.
			// (2) Firmware version of the drive 1 is different from the supported version.
			// (3) Firmware version of the drive 2 is different from the supported version.
			// (4) Firmware version of the printer is not supported.
			msg += tr("L7008:\n"
				"The output device is an unsupported version.\n"
				"You cannot publish to this output device.\n"
				"For details, please contact support.");
			break;

		default:
			msg += tr("Unknown INFOMATION Code");
			break;
		}
	}

	if ( msg.isEmpty() && !jobStatusPublishing.empty() )
	{
		for(const auto& jobStatus : jobStatusPublishing)
		{
			if( jobStatus.ulJobStatus == PP100API_JOB_PAUSED ||
				jobStatus.ulJobStatus == PP100API_JOB_WAITING ||
				jobStatus.ulJobStatus == PP100API_JOB_STANDBY )
			{
				switch(jobStatus.ulErrorCode)
				{
				case PP100API_STP000: msg += "PP100API_STP000"; break;
				case PP100API_STP001: msg += "PP100API_STP001"; break;
				case PP100API_STP002: msg += "PP100API_STP002"; break;
				case PP100API_RTN000: msg += "PP100API_RTN000"; break;
				case PP100API_RTN001: msg += "PP100API_RTN001"; break;
				case PP100API_RTN002: msg += "PP100API_RTN002"; break;
				case PP100API_RTN003: msg += "PP100API_RTN003"; break;
				case PP100API_RTN004: msg += "PP100API_RTN004"; break;
				case PP100API_RTN005: msg += "PP100API_RTN005"; break;
				case PP100API_RTN006: msg += "PP100API_RTN006"; break;
				case PP100API_RTN007: msg += "PP100API_RTN007"; break;
				case PP100API_RTN008: msg += "PP100API_RTN008"; break;
				case PP100API_RTN009: msg += "PP100API_RTN009"; break;
				case PP100API_RTN010: msg += "PP100API_RTN010"; break;
				case PP100API_RTN011: msg += "PP100API_RTN011"; break;
				case PP100API_RTN012: msg += "PP100API_RTN012"; break;
				case PP100API_RTN013: msg += "PP100API_RTN013"; break;
				case PP100API_RTN014: msg += "PP100API_RTN014"; break;
				case PP100API_RTN015: msg += "PP100API_RTN015"; break;
				case PP100API_RTN016: msg += "PP100API_RTN016"; break;
				case PP100API_RTN017: msg += "PP100API_RTN017"; break;
				case PP100API_SUCCESS: break;
				default:
					 msg += QString::asprintf("%lu", jobStatus.ulErrorCode);
					 break;
				}
				break;
			}
		}
	}

	if ( msg.isEmpty() && !jobStatusPublishing.empty() )
	{
		switch(jobStatusPublishing[0].ulJobStatus)
		{
		case PP100API_JOB_PAUSED:  msg = "PP100API_JOB_PAUSED";  break;
		case PP100API_JOB_WAITING: msg = "PP100API_JOB_WAITING"; break;
		case PP100API_JOB_STANDBY: msg = "PP100API_JOB_STANDBY"; break;
		default: break;
		}
	}

	if ( msg.isEmpty() )
	{
		for(const auto& jobStatus : jobStatusPublishing)
		{
			if(
				jobStatus.ulDestination == PP100API_STACKER4 ||
				jobStatus.ulDestination == PP100API_STACKER2_OR_STACKER4 )
			{
				msg = tr("Publishing is possible.\n\nDo not pull out Stacker 4 while discs are being transferred to it.");
				break;
			}
		}
	}

	if ( msg.isEmpty() )
	{
		msg = tr("You can now publish discs.");
	}

	textDeviceStatus->setText(msg);
}


// Drive status setting
void TDMonitorEx::setDriveInfo(const PP100_PUBLISHER_STATUS& publisherStatus, int drvNum)
{
	QProgressBar *progressBar_DriveN = nullptr;
	QLabel *label_DriveNStatus = nullptr;
	if(drvNum == 0)
	{
		progressBar_DriveN = progressBar_Drive1;
		label_DriveNStatus = label_Drive1Status;
	}
	else if(drvNum == 1)
	{
		progressBar_DriveN = progressBar_Drive2;
		label_DriveNStatus = label_Drive2Status;
	}
	else
	{
		assert(false);
		return;
	}

	QString strStyleWrite = 
		"QProgressBar {" LF
		"	border: 1px solid gray;" LF
		"	border-radius: 4px;" LF
		"}" LF
		"" LF
		"QProgressBar::chunk {" LF
		"	border-right: 0.5px solid #258;" LF
		"	background-color: #5AF;" LF
		"}" LF;

	QString strStyleCompare = 
		"QProgressBar {" LF
		"	border: 1px solid gray;" LF
		"	border-radius: 4px;" LF
		"}" LF
		"" LF
		"QProgressBar::chunk {" LF
		"	border-right: 0.5px solid #070;" LF
		"	background-color: #0E0;" LF
		"}" LF;

	QString msg;
	switch(publisherStatus.ulDriveStatus[drvNum])
	{
	case PP100API_UNUSED:
		progressBar_DriveN->setStyleSheet(strStyleWrite);
		msg = tr("Unused");
		break;
	case PP100API_TRANSFERRING:
		progressBar_DriveN->setStyleSheet(strStyleWrite);
		msg = tr("Transferring");
		break;
	case PP100API_PAUSED:
		msg = tr("Stopping");
		break;
	case PP100API_WRITING:
		progressBar_DriveN->setStyleSheet(strStyleWrite);
		msg = tr("Writing");
		break;
	case PP100API_VERIFYING_WRITE:
		progressBar_DriveN->setStyleSheet(strStyleCompare);
		msg = tr("Verifying Write");
		break;
	case PP100API_ERROR:
		msg = tr("Error");
		break;
	case PP100API_WAITING:
		progressBar_DriveN->setStyleSheet(strStyleWrite);
		msg = tr("Standby");
		break;
	default:
		fprintf(stderr, "unknown drive status: %lu\n", publisherStatus.ulDriveStatus[drvNum]);
		msg = tr("Unknown");
		break;
	}

	label_DriveNStatus->setText(msg);
	if ( publisherStatus.ulDriveProgress[drvNum] != 0 )
	{
		progressBar_DriveN->setValue(publisherStatus.ulDriveProgress[drvNum]);
	}
	else
	{
		progressBar_DriveN->reset();
	}
}


// Stacker status setting
void TDMonitorEx::setStackerSetting(const PP100_PUBLISHER_STATUS& publisherStatus, int stckerNum)
{
	QString msg;
	bool isUse = true;

	QLabel * labelStatus = NULL;
	QLabel * labelRest = NULL;
	switch(stckerNum)
	{
	case 0:		// Stacker1
		labelStatus = labelStacker1Status;
		labelRest = labelStacker1Rest;
		break;

	case 1:		// Stacker2
		labelStatus = labelStacker2Status;
		labelRest = labelStacker2Rest;
		break;

	case 2:		// Stacker3
		labelStatus = labelStacker3Status;
		labelRest = labelStacker3Rest;
		break;

	case 3:		// Stacker4
		labelStatus = labelStacker4Status;
		labelRest = labelStacker4Rest;
		break;

	default:
		assert(false);
		break;
	}

	if ( labelStatus != NULL )
	{
		QString status;
		switch(publisherStatus.ulStackerSetting[stckerNum])
		{
		case PP100API_INPUT:            status = tr("Input Stacker");  break;
		case PP100API_OUTPUT:           status = tr("Output Stacker"); break;
		case PP100API_NOT_USE_STACKER:  status = tr("Unused");         isUse = false; break;
		case PP100API_CD_MINUS_R:       status = tr("CD-R");           break;
		case PP100API_DVD_MINUS_R:      // obsolated
		case PP100API_DVD_PLUS_R:       // obsolated
		case PP100API_DVD_R:            status = tr("DVD(+R/-R)");     break;
		case PP100API_DVD_MINUS_R_DL:   // obsolated
		case PP100API_DVD_PLUS_R_DL:    // obsolated
		case PP100API_DVD_R_DL:         status = tr("DVD(+R/-R) DL");  break;
		case PP100API_BD_MINUS_R:       status = tr("BD-R");           break;
		case PP100API_BD_MINUS_R_DL:    status = tr("BD-R DL");        break;
		case PP100API_BD_MINUS_R_XL100: status = tr("BD-R XL 100");    break;
		case PP100API_BD_MINUS_R_XL128: status = tr("BD-R XL 128");    break;
		case PP100API_ERROR:
			fprintf(stderr, "Wrong stacker setting: PP100API_ERROR\n");
			status = tr("Unknown");
			break;
		default:
			fprintf(stderr, "UNKNOWN stacker setting: %lu\n", publisherStatus.ulStackerSetting[stckerNum]);
			status = tr("Unknown");
			break;
		}
		labelStatus->setText(status);
	}

	if ( labelRest != NULL )
	{
		// Stacker Rest
		labelRest->setStyleSheet("");
		if(isUse)
		{
			switch(stckerNum)
			{
			case 0:		// Stacker1
			case 1:		// Stacker2
				switch( publisherStatus.ulStackerRest[stckerNum] )
				{
				case PP100API_ERROR:
					labelRest->setStyleSheet("background-color : rgb(255,96,0); color:white; font-weight:bold; border-radius:8px;");
					labelRest->setText("-");
					break;

				case PP100API_STACKER_NOT_FULL:
					labelRest->setText("-");
					break;

				case 100:
					labelRest->setStyleSheet("background-color : rgb(255,96,0); color:white; font-weight:bold; border-radius:8px;");
					labelRest->setText("100 %");
					break;

				default:
					labelRest->setText(QString::asprintf("%lu %%", publisherStatus.ulStackerRest[stckerNum]));
					break;
				}
				break;

			case 2:		// Stacker3
			case 3:		// Stacker4
				switch( publisherStatus.ulStackerRest[stckerNum] )
				{
				case 100:
					labelRest->setStyleSheet("background-color : rgb(255,96,0); color:white; font-weight:bold; border-radius:8px;");
					labelRest->setText("100 %");
					break;

				case 0:
					labelRest->setText("0 %");
					break;

				default:
					labelRest->setText("-");
				}
				break;

			default:
				assert(false);
				break;
			}
		}
		else
		{
			labelRest->setText("-");
		}
	}
}


// JOB cancel processing
void TDMonitorEx::cancelJob()
{
	// Get selected item.
	std::vector<std::string> jobIdList;
	for(auto item : treeJobList->selectedItems())
	{
		jobIdList.push_back( item->data(colJobID, Qt::UserRole).toString().toStdString() );
	}

	if (!jobIdList.empty())
	{
		bool failed = false;
		for(auto jobId : jobIdList)
		{
			char szJobID[41] = {0};
			strcpy(szJobID, jobId.c_str());
			long ret = g_Api.PP100_CancelJob(g_Api.ulHandle, szJobID);
			if(PP100API_SUCCESS != ret)
			{
				failed = true;
				// Cancel until the end even if there is an error.
				fprintf(stderr, "PP100_CancelJob(,\"%s\") returns %ld\n", szJobID, ret);
			}
		}
		if ( failed )
		{
			(void) QMessageBox::warning(
				NULL,
				QApplication::translate("main function", "EPSON Total Disc Monitor"),
				QApplication::translate("main function", "Some jobs could not be performed."));
		}
		fetchStatus(); // -> updateScreenSignal() -> updateScreen()
		m_timer->start();
	}
}


// JOB all selection processing
void TDMonitorEx::selectAllJobs()
{
	for(int nCnt = 0; nCnt < treeJobList->topLevelItemCount(); ++nCnt)
	{
		QTreeWidgetItem *item = treeJobList->topLevelItem(nCnt);
		item->setSelected(true);
	}
}


// Unfinished JOB tab processing
void TDMonitorEx::showUnfinishedJobs()
{
	tabWidget->setCurrentIndex(0);
	QTreeWidgetItem *item = treeJobList->headerItem();
	item->setText(colUnfinishedCopies,     tr("Copies"));
	item->setText(colUnfinishedJobType,    tr("Type"));
	item->setText(colUnfinishedStackerIn,  tr("Input Stacker"));
	item->setText(colUnfinishedStackerOut, tr("Output Stacker"));
	item->setText(colUnfinishedRemaining,  tr("Time Needed"));
	treeJobList->setContextMenuPolicy(Qt::ActionsContextMenu);
	actionShowUnfinishedJobs->setChecked(true);
	updateScreen();
	m_timer->start();
}


// Finished JOB tab processing
void TDMonitorEx::showFinishedJobs()
{
	tabWidget->setCurrentIndex(1);
	QTreeWidgetItem *item = treeJobList->headerItem();
	item->setText(colFinishedComletionNumber,   tr("Successful"));
	item->setText(colFinishedErrorNumber,       tr("Failed"));
	item->setText(colFinishedPublicationNumber, tr("Quantity"));
	item->setText(colFinishedJobType,           tr("Type"));
	item->setText(colFinishedStackerOut,        tr("Output Stacker"));
	treeJobList->setContextMenuPolicy(Qt::PreventContextMenu);
	actionShowFinishedJobs->setChecked(true);
	updateScreen();
	m_timer->start();
}


// Display version information dialog
void TDMonitorEx::showAboutDialog()
{
	QDialog *dialog = new QDialog;
	Ui::AboutDialog ui;
	ui.setupUi(dialog);
	dialog->setModal(true);
	dialog->show();
	(void) dialog->exec();
}


// Tab switching process
void TDMonitorEx::changeTab(int num)
{
	if(!num)
	{
		showUnfinishedJobs();
	}
	else
	{
		showFinishedJobs();
	}
}


// Screen update process
void TDMonitorEx::updateStatusAsync()
{
	if ( m_mtxFetchStatus.tryLock() )
	{
		QThreadPool::globalInstance()->start( new MyRunnable([&] (void) -> void {
			fetchStatus();
		}) );
		m_mtxFetchStatus.unlock();
	}
}


// Screen update process
void TDMonitorEx::updateScreen()
{
	updateScreen_JobList();
	updateScreen_DeviceStatus();
	updateMenu();
	repaint();
}


// toggle minimaized, toggle dark mode.
bool TDMonitorEx::event(QEvent* event)
{
	if(event->type() == QEvent::WindowStateChange)
	{
		// Screen update cycle setting (normal time: 5000msec, window minimization: stop)
		QWindowStateChangeEvent* ev = static_cast<QWindowStateChangeEvent*>(event);
		if(!(ev->oldState() & Qt::WindowMinimized) && (windowState() & Qt::WindowMinimized))
		{
			m_timer->stop();
		}
		else if((ev->oldState() & Qt::WindowMinimized) && !(windowState() & Qt::WindowMinimized))
		{
			m_timer->setInterval(5000);
			fetchStatus(); // -> updateScreenSignal() -> updateScreen()
			m_timer->start();
		}
	}
	if(event->type() == QEvent::ApplicationPaletteChange)
	{
		textDeviceStatus->setStyleSheet(textDeviceStatus->styleSheet());
		frameInkStatus->setStyleSheet(frameInkStatus->styleSheet());
		frameDrive1Status->setStyleSheet(frameDrive1Status->styleSheet());
		frameDrive2Status->setStyleSheet(frameDrive2Status->styleSheet());
		framePrinterStatus->setStyleSheet(framePrinterStatus->styleSheet());
		frameStackerStatus->setStyleSheet(frameStackerStatus->styleSheet());
	}
	return QMainWindow::event(event);
}


// Menu enable / disable update
void TDMonitorEx::updateMenu()
{
	if(treeJobList->topLevelItemCount() != 0 && tabWidget->currentIndex() == 0)
	{
		QList<QTreeWidgetItem *> itemList = treeJobList->selectedItems();

		int cancelable = 0;
		for( auto & item : itemList )
		{
			// jobStatus.ulJobType
			switch((unsigned long) item->data(colUnfinishedJobType, Qt::UserRole).toULongLong())
			{
			case PP100API_JOB_TYPE_PUBLISH:
			case PP100API_JOB_TYPE_WRITE:
			case PP100API_JOB_TYPE_PRINT:
			case PP100API_JOB_TYPE_MEASURE:
				cancelable++;
				break;
			default:
				break;
			}
		}

		// if selected items include cancelable job, enable cancel action.
		if(cancelable > 0)
		{
			actionCancelJob->setEnabled(true);
		}
		else
		{
			actionCancelJob->setEnabled(false);
		}
		actionSelectAllJobs->setEnabled(true);
	}
	else
	{
		actionCancelJob->setEnabled(false);
		actionSelectAllJobs->setEnabled(false);
	}
}
